#!/usr/bin/env python3
# -*- coding: utf-8 -*-
'''
Routines for analysis

:copyright:
    Samuel Weber (samuel.weber@geo.uzh.ch)
:version:
    April 9, 2018
'''


import yaml, os
import fnmatch

# load config file
filename = 'config.yaml'
if os.path.isfile(filename):
    with open(filename, 'r') as f:
        config = yaml.load(f)
else:
    raise IOError("Parameter file not found [%s]" % filename)
    

def get_user_host():
    import socket
    import getpass
    username = getpass.getuser()
    hostname = socket.gethostname()
    return username, hostname


def get_data_path(station):
    user, host = get_user_host()
    for u_id in config['users']:
        u = config['users'][u_id]
        # loop through users in config file and check if current user
        # has a valid entry. fnmatch allows for wildcards (e.g. '*')
        if fnmatch.fnmatch(user,u['name']) and fnmatch.fnmatch(host,u['host']):
            if station in ['MH36', 'MH38']:
                return u['permasense_vault_path'] + 'geophones/binaries/PS/'
            elif station == 'MH40':
                return u['permasense_ims_path'] + 'data/binaries/PS/'
            elif station == 'IMIS':
                return u['permasense_vault_path'] + 'data_archive/meteo_data/IMIS/'
            else:
                print('No valid station name')

    print('Add your user with correct datapaths to config.yaml')


def get_source_path():
    user, host = get_user_host()
    for u_id in config['users']:
        u = config['users'][u_id]
        # loop through users in config file and check if current user
        # has a valid entry. fnmatch allows for wildcards (e.g. '*')
        if fnmatch.fnmatch(user,u['name']) and fnmatch.fnmatch(host,u['host']):
            return u['source_path']

    print('Add your user with correct source_path to config.yaml')

    
def get_target_path():
    user, host = get_user_host()
    for u_id in config['users']:
        u = config['users'][u_id]
        # loop through users in config file and check if current user
        # has a valid entry. fnmatch allows for wildcards (e.g. '*')
        if fnmatch.fnmatch(user,u['name']) and fnmatch.fnmatch(host,u['host']):
            return u['target_path']
            
    print('Add your user with correct target_path to config.yaml')


def get_vault_path():
    user, host = get_user_host()
    for u_id in config['users']:
        u = config['users'][u_id]
        # loop through users in config file and check if current user
        # has a valid entry. fnmatch allows for wildcards (e.g. '*')
        if fnmatch.fnmatch(user,u['name']) and fnmatch.fnmatch(host,u['host']):
            return u['permasense_vault_path']
            
    print('Add your user with correct permasense_vault_path to config.yaml')


def station2stat(station):
    if station == 'MH36':
        return 'MHDL'
    elif station == 'MH38':
        return 'MHDT'
    elif station == 'MH40':
        return 'MH40'

        
def stat2station(stat):
    if stat == 'MHDL':
        return 'MH36'
    elif stat == 'MHDT':
        return 'MH38'
    elif stat == 'MH40':
        return 'MH40'


def yyyymmddhh(time):
    from obspy import UTCDateTime
    t = UTCDateTime(time)
        
    return t.datetime.strftime('%Y%m%d%H')


def yyyymmdd_hhmmss(time):
    from obspy import UTCDateTime
    t = UTCDateTime(time)
            
    return t.datetime.strftime('%Y%m%d_%H%M%S')

    
def yyyymmdd_hhmmss_sss(time):
    from obspy import UTCDateTime
    t = UTCDateTime(str(time))

    return t.datetime.strftime('%Y%m%d_%H%M%S_%f')[:-3]


def import_miniseed_data(timewindow, station, channels):
    from obspy.core import read, Stream, UTCDateTime
    import numpy as np
    st_list = Stream()
    path_data = get_data_path(station)
    stat = station2stat(station)
    
    if type(timewindow) is UTCDateTime:
        t1 = timewindow
        timestamp = [UTCDateTime(t1.year,t1.month,t1.day,t1.hour)]
        t2 = t1 + 3600
    elif type(timewindow) is list:
        t1 = timewindow[0]
        t2 = timewindow[-1]
        if t1 == t2:
            timestamp = [UTCDateTime(t1.year,t1.month,t1.day,t1.hour)]
        elif (t2.minute == 0) and (t2 > t1):
            timestamp = np.arange(UTCDateTime(t1.year,t1.month,t1.day,t1.hour),UTCDateTime(t2.year,t2.month,t2.day,t2.hour), 3600)
        elif (t2.minute > 0) and (t2 > t1):
            timestamp = np.arange(UTCDateTime(t1.year,t1.month,t1.day,t1.hour),UTCDateTime(t2.year,t2.month,t2.day,t2.hour+1), 3600)
        else:
            print('This scenario is not defined')
    
    for time in timestamp:
        for channel in channels:
            fname = path_data + station + '/' + str(UTCDateTime(time).year) + \
                '/' + channel + '.D/PS.' + stat + '.A.' + channel + '.D.' + \
                yyyymmdd_hhmmss(time) + '.miniseed'
                
            try:
                st_tmp = read(fname)
                st_list += st_tmp
            except:
                print('... no file with name ' + fname)
    
    st = st_list.merge(method=0, fill_value=0)
    st = st.slice(t1,t2-st[0].stats.delta)
    print('The imported data:')
    print(st)
    return st


def calc_spectrum(trace, dBscale=False):
    import numpy as np
    freq = np.fft.rfftfreq(len(trace), trace.stats.delta)
    npts = trace.stats.npts
    if dBscale:
        power = 10 * np.log10(np.abs(np.fft.rfft(trace)/npts))
    else:
        power = np.abs(np.fft.rfft(trace)/npts)    # fft computing and normalization
    return freq, power


def gen_df_waveform(trace, TBEG, TEND):
    import numpy as np
    import pandas as pd
    import datetime as dt
    if TBEG < trace.stats['starttime']:
        TBEG = trace.stats['starttime']
    if TEND > trace.stats['endtime']:
        TEND = trace.stats['endtime']
    timestamp2datetime = np.vectorize(dt.datetime.utcfromtimestamp)    
    timestamp = np.arange(TBEG.timestamp,TEND.timestamp + trace.stats['delta'],trace.stats['delta'])
    tr = trace.slice(TBEG,TEND)
    time = timestamp2datetime(timestamp)[:len(tr.data)]
    waveform = pd.DataFrame({'time':time, 'wave':tr.data}, columns=['time','wave'])
    waveform = waveform.set_index('time')
    return waveform

    
def conv_data2velocity(stream, remove_response=True):
    import numpy as np
    from obspy import read_inventory
    station = stat2station(stream[0].stats.station)
    if station in ['MH36','MH38']:
        if remove_response:
            stat = station2stat(station)
            inv = read_inventory(stat + '.xml')
            stream.remove_response(inventory=inv, output='VEL', water_level = 0)
        else:
            for i in range(len(stream)):
                stream[i].data = stream[i].data / (6.4*10**8)
    elif station in ['MH40']:
        for i in range(len(stream)):
            fs = stream[i].stats.sampling_rate
            stream[i].data = (stream[i].data[:-1] + np.diff(stream[i].data)/2) * fs**-1
    return stream


def detrend_stream(stream, FREQMIN=-1, FREQMAX=-1):
    import numpy as np
    for i in range(len(stream)):
        stream[i].data = stream[i].data - np.nanmean(stream[i].data)
    return stream


def filter_stream(stream, FREQMIN=-1, FREQMAX=-1, corners=4):
    import sys
    from trigger_events_miniseed import get_stalta_setting
    station = stat2station(stream[0].stats.station)
    if (FREQMIN < 0 or FREQMAX < 0):
        COEFUP, COEFDOWN, SUMSTA, STA, LTA, FREQMIN, FREQMAX = get_stalta_setting(station)
    elif FREQMIN >= FREQMAX:
        print('FREQMAX must be bigger than FREQMIN')
        sys.exit(0)
    if FREQMAX > stream[0].stats['sampling_rate'] / 2:
        stream = stream.filter('highpass', freq=FREQMIN)
    else:
        stream = stream.filter('bandpass',freqmin=FREQMIN, freqmax=FREQMAX,corners=corners)
    return stream


def spectrogram(trace, freq_lim=[0,0], window=1, overlap=0.95, mult=8, dB=True):
    print('spectrogram() returns spectrum (\'psd\'), frequency and time')
    import numpy as np
    import datetime as dt
    timestamp2datetime = np.vectorize(dt.datetime.utcfromtimestamp)
    import matplotlib as mpl
    from trigger_events_miniseed import get_stalta_setting
    from matplotlib.colors import Normalize
    
    if sum(freq_lim) == 0:
        station = stat2station(trace.stats.station)
        COEFUP, COEFDOWN, SUMSTA, STA, LTA, FREQMIN, FREQMAX = get_stalta_setting(station)
        freq_lim = [FREQMIN, FREQMAX]
    fs = int(trace.stats.sampling_rate)
    nfft = int(pow(2, np.ceil(np.log2(window * fs))))
    if mult is not None:
        mult = int(pow(2, np.ceil(np.log2(mult))))
        mult = mult * nfft
    noverlap = int(overlap * nfft)
    data = trace.data - np.nanmean(trace.data)
    print('pad_to=mult is missing')
    S, f, t = mpl.mlab.specgram(data, Fs=fs, NFFT=nfft, noverlap=noverlap)
    if dB==True:
        S = 10 * np.log10(S[1:, :]) # Log scaling for amplitude values, factor 20 for dB
    else:
        S = np.sqrt(S[1:, :])
    S = np.ma.masked_where(np.isnan(S),S)
    f = f[1:]
    
    S = S[(f >= freq_lim[0]) & (f <= freq_lim[1])]
    f = f[(f >= freq_lim[0]) & (f <= freq_lim[1])]
    
    norm = Normalize(np.min(S), np.max(S), clip=True)
                        
    timestamp = trace.stats.starttime.timestamp + t
    time = timestamp2datetime(timestamp)
    return S, f, time, norm


def crosscorr(stream, channel, stream_template, channel_template):
    import numpy as np
    import datetime as dt
    timestamp2datetime = np.vectorize(dt.datetime.utcfromtimestamp)
    from xcor import xcorr
    if stream[0].stats.sampling_rate > stream_template[0].stats.sampling_rate:
        print(stat2station(stream[0].stats.station) + ' get resampled to ' + str(stream_template[0].stats.sampling_rate))
        stream.resample(stream_template[0].stats.sampling_rate)
    elif stream[0].stats.sampling_rate < stream_template[0].stats.sampling_rate:
        print(stat2station(stream_template[0].stats.station) + ' get resampled to ' + str(stream[0].stats.sampling_rate))
        stream_template.resample(stream[0].stats.sampling_rate)
    else:
        print('No resampling needed')
    
    trace = stream.select(channel=channel)[0]
    trace_template = stream_template.select(channel=channel_template)[0]
    corr = xcorr(trace.data, trace_template.data)
    time_shift = (len(trace.data) - corr.shape[0]) / trace.stats.sampling_rate * 0.5
    timestamp = np.linspace(trace.stats.starttime.timestamp + time_shift, trace.stats.endtime.timestamp - time_shift, corr.shape[0])
    time = timestamp2datetime(timestamp)
    return corr, time


def nan2zeroStream(stream):
    import numpy as np
    for i in range(len(stream)):
        stream[i].data = np.nan_to_num(stream[i].data)
    return stream


def detectCoincidence(event_df1, event_df2, amp_level1 = 0, amp_level2 = 0, tol = 0.1):
    try:
        A = event_df1[event_df1.amp_max >= amp_level1]
        B = event_df2[event_df2.amp_max >= amp_level2]
        print('Cutoff at amplitude levels ' + str(amp_level1) + ' and ' + str(amp_level2))
    except:
        A = event_df1
        B = event_df2
        print('Variable amp_max is not given.')
    
    AA = A[~A.index.duplicated(keep='first')]
    BB = B[~B.index.duplicated(keep='first')]
    
    AAA = AA.reindex(BB.index, method='nearest',tolerance=str(tol)+'Sec')
    AAA = AAA.reindex(AA.index, method='nearest',tolerance=str(tol)+'Sec')
    try:
        AAAA = AAA[AAA.amp_max > 0]
    except:
        AAAA = AAA[AAA.duration > 0]
        print('WARNING: \'DataFrame\' object has no attribute \'amp_max\'')
    AAAAA = AAAA.sort_index()
    
    BBB = BB.reindex(AA.index, method='nearest',tolerance=str(tol)+'Sec')
    try:
        BBBB = BBB[BBB.amp_max > 0]
    except:
        BBBB = BBB[BBB.duration > 0]
        print('WARNING: \'DataFrame\' object has no attribute \'amp_max\'')
    BBBBB = BBBB.sort_index()
    
    print('Done. Returns two pandas dataframe with the coincidentally triggered events.')
    return AAAAA, BBBBB



